@@ -182,7 +182,7 @@ void CExpNet::_release() {
 
 	release_pqcC();
 	// First release CExpNet's resources then its parents CATNET
-	CATNET<char, MAX_NODE_NAME, double>::_release();
+	CATNET<char, double>::_release();
 }
 
 void CExpNet::release_pqcC() {
@@ -201,7 +201,7 @@ void CExpNet::release_pqcC() {
 }
 
 void CExpNet::_reset() {
-	CATNET<char, MAX_NODE_NAME, double>::_reset();
+	CATNET<char, double>::_reset();
 	m_lambdas = 0;
 	m_loglambdas = 0;
 	m_curNode =0;
@@ -215,7 +215,7 @@ void CExpNet::_reset() {
 
 CExpNet& CExpNet::operator =(const CExpNet &cnet) {
 
-	CATNET<char, MAX_NODE_NAME, double> ::init(
+	CATNET<char, double> ::init(
 			(int)cnet.m_numNodes, (int)cnet.m_maxParents, (int)cnet.m_maxCategories,
 			(const char**)cnet.m_nodeNames, (const int*)cnet.m_numParents, 
 			(const int**)cnet.m_parents,
@@ -726,7 +726,15 @@ int* CExpNet::catnetSample(int nsamples) {
 				pnodesample[i] = (int)m_pCatnetSamples[j * m_numNodes + pnodepars[i]];
 			}
 			pnodeprob = pProbList->find_slot(0, pnodesample, 0);
-
+			if(!pnodeprob) {
+				if(m_pCatnetSamples)
+					CATNET_FREE(m_pCatnetSamples);
+				m_pCatnetSamples = 0;
+				m_nCatnetSamples = 0;
+				CATNET_FREE(pnodesample);
+				CATNET_FREE(porder);
+				return 0;
+			}
 			u = (double)rand() / (double)RAND_MAX;
 			v = 0;
 			for(i = 0; i < m_numCategories[nnode]; i++) {
@@ -822,6 +830,9 @@ int CExpNet::predict(double *psamples, int nsamples) {
 		return ERR_CATNET_PROC;
 
 	for(k = 0; k < m_numNodes; k++) {
+		/* work with the sample distribution of pCurNet if necessary */
+		set_sample_cache();
+
 		nnode = porder[k];
 		pProbList = (PROB_LIST<double>*)getNodeProb(nnode);
 		pnodeprob = pProbList->pProbs;
@@ -933,7 +944,7 @@ double CExpNet::findLogNodeLikelihood(int nnode, double *psamples, int nsamples)
 double CExpNet::findNodeLoglik(int nnode, double *psamples, int nsamples) {
 	
 	double faux, fLogLik;
-	int j, cC_setSize, ic, icC, iC;
+	int res, 	j, cC_setSize, ic, icC, iC;
 	double *pnodeprob;
 	PROB_LIST<double>* pProbList;
 	
@@ -954,7 +965,14 @@ double CExpNet::findNodeLoglik(int nnode, double *psamples, int nsamples) {
 		cC_setSize = pcC_size();
 	}
 	else {
-		findNodeMarginalProb(nnode, psamples, nsamples);
+		res = findNodeMarginalProb(nnode, psamples, nsamples);
+		if(res == ERR_CATNET_MEM) {
+			int nsize = (int)exp(log((double)(maxCategories()))*(1+maxParentSet()))*1000;
+			catnetSample(nsize);
+			res = findNodeMarginalProb(nnode, psamples, nsamples);
+		}
+		if(res != ERR_CATNET_OK)
+			return -FLT_MAX;
 	}
 	
 	fLogLik = 0;
@@ -987,3 +1005,225 @@ double CExpNet::findNodeLoglik(int nnode, double *psamples, int nsamples) {
 	return fLogLik;
 }
 
+int CExpNet::estimateNodeProb(int nnode, int *parnodes, int numpars, double *psamples, int nsamples) {
+
+	int res, i, j, ic, iC;
+	int *pcats, numcats;
+	double fdiff, fdiffsum, fsum, *paux;
+	double *pnodeprob;
+	PROB_LIST<double>* pProbList;
+
+	if(nnode < 0 || nnode >= m_numNodes || !m_loglambdas || 
+		!psamples || nsamples < 1)
+		return ERR_CATNET_PARAM;
+		
+	pProbList = (PROB_LIST<double>*)getNodeProb(nnode);
+	pnodeprob = pProbList->pProbs;
+
+	numcats = m_numCategories[nnode];
+
+	if(numpars <= 0 || !parnodes) {
+		fsum = 0;
+		for(ic = 0; ic < numcats; ic++) {
+			fdiffsum = 0;
+			for(j = 0; j < nsamples; j++) {
+				fdiff = (psamples[j * m_numNodes + nnode] * m_loglambdas[nnode][numcats+ic] + m_loglambdas[nnode][ic]);
+				// P(Y_i|X_i=c)
+				fdiffsum += exp(-fdiff);
+			}
+			pnodeprob[ic] = fdiffsum;
+			fsum += fdiffsum;
+		}
+		if(fsum > 0)
+			fsum = 1/fsum;
+		for(ic = 0; ic < numcats; ic++)
+			pnodeprob[ic] *= fsum;
+		return ERR_CATNET_OK;
+	}
+
+	// at this point m_nBlockSizes == numnodes and m_pBlockSizes are filled
+	m_parCatSetSize = 1;
+	for(i = 0; i < numpars; i++) {
+		m_parCatSetSize *= m_numCategories[parnodes[i]];
+	}
+
+	if(m_pcC) 
+		CATNET_FREE(m_pcC);
+	m_pcC = 0;
+	if(m_qcC) 
+		CATNET_FREE(m_qcC);
+	m_qcC = 0;
+
+	// m_pcC returns marginal P(X_{Pa_i}) 
+	m_pcC = (double*)CATNET_MALLOC(m_parCatSetSize*sizeof(double));
+	pcats = (int*)CATNET_MALLOC(m_parCatSetSize*numpars*sizeof(int));
+	if(!m_pcC || !pcats)
+		return ERR_CATNET_MEM;
+	memset(m_pcC, 0, m_parCatSetSize*sizeof(double));
+
+	if(!m_pCatnetSamples || m_nCatnetSamples < 1) {
+		// find the exact joint probability for small sets
+		res = marginalProb(parnodes, numpars);
+		if(res != ERR_CATNET_OK) {
+			CATNET_FREE(pcats);
+			return res;
+		}
+		if(m_margProbSize != m_parCatSetSize) {
+			CATNET_FREE(pcats);
+			return ERR_CATNET_PROC;
+		}
+		for(ic = 0; ic < m_parCatSetSize; ic++) {
+			// determine the category indices
+			fsum = ic;
+			for(i = 0; i < numpars; i++) {
+				pcats[ic * numpars + i] = (int)(fsum/m_pBlockSizes[i]);
+				fsum -= pcats[ic*numpars + i] * m_pBlockSizes[i];
+			}
+			m_pcC[ic] = getCatProb(pcats + ic * numpars, numpars);
+		}
+	}
+	else {
+		if(!m_pBlockSizes || m_nBlockSizes < numpars) {
+			if(m_pBlockSizes)
+				CATNET_FREE(m_pBlockSizes);
+			m_nBlockSizes = numpars;
+			m_pBlockSizes = (int*) CATNET_MALLOC(m_nBlockSizes * sizeof(int));
+		}
+		m_pBlockSizes[numpars - 1] = 1;
+		for (i = numpars - 2; i >= 0; i--) {
+			m_pBlockSizes[i] = m_pBlockSizes[i + 1] * m_numCategories[parnodes[i + 1]];
+			
+		}
+		for(j = 0; j < m_nCatnetSamples; j++) {
+			ic = 0;
+			for (i = 0; i < numpars; i++)
+				ic += (m_pBlockSizes[i] * m_pCatnetSamples[j * m_numNodes + parnodes[i]]);
+			m_pcC[ic] += 1;
+		}
+		fsum = 1/(double)m_nCatnetSamples;
+		for(ic = 0; ic < m_parCatSetSize; ic++)
+			m_pcC[ic] *= fsum;
+
+		for(ic = 0; ic < m_parCatSetSize; ic++) {
+			// determine the category indices
+			fsum = ic;
+			for(i = 0; i < numpars; i++) {
+				pcats[ic * numpars + i] = (int)(fsum/m_pBlockSizes[i]);
+				fsum -= pcats[ic*numpars + i] * m_pBlockSizes[i];
+			}
+		}
+	}
+
+	paux = (double*)CATNET_MALLOC(m_parCatSetSize*sizeof(double));
+	m_qcC = (double*)CATNET_MALLOC(m_parCatSetSize*nsamples*sizeof(double));
+
+	for (j = 0; j < nsamples; j++) {
+		fsum = 0;
+		for(ic = 0; ic < m_parCatSetSize; ic++) {
+			fdiffsum = 0;
+			for(i = 0; i < numpars; i++) {
+				fdiffsum += (psamples[j * m_numNodes + parnodes[i]] * m_loglambdas[parnodes[i]][m_numCategories[parnodes[i]]+pcats[ic*numpars + i]] + m_loglambdas[parnodes[i]][pcats[ic*numpars + i]]);
+			}
+			paux[ic] = m_pcC[ic]*exp(-fdiffsum);
+			fsum += paux[ic];
+		}
+		if(fsum > 0) 
+			fsum = 1/fsum;
+		else
+			fsum = FLT_MAX;
+		// set P(X_{Pa_i}|Y_{Pa_i}) in m_qcC
+		for(ic = 0; ic < m_parCatSetSize; ic++)
+			m_qcC[j*m_parCatSetSize + ic] = fsum*paux[ic];
+	}
+
+	for(iC = 0; iC < m_parCatSetSize; iC++) {
+		fsum = 0;
+		for(ic = 0; ic < numcats; ic++) {
+			fdiffsum = 0;
+			for(j = 0; j < nsamples; j++) {
+				fdiff = (psamples[j * m_numNodes + nnode] * m_loglambdas[nnode][numcats+ic] + m_loglambdas[nnode][ic]);
+				// P(Y_i|X_i=c)P(Pa_i=C|Y_{Pa_i})
+				fdiffsum += exp(-fdiff)*m_qcC[j*m_parCatSetSize + iC];
+			}
+			pnodeprob[iC*numcats + ic] = fdiffsum;
+			fsum += fdiffsum;
+		}
+		if(fsum > 0)
+			fsum = 1/fsum;
+		for(ic = 0; ic < numcats; ic++)
+			pnodeprob[iC*numcats + ic] *= fsum;
+	}
+
+	CATNET_FREE(pcats);
+	CATNET_FREE(paux);
+
+	// we don't need them any more
+	if(m_pcC) 
+		CATNET_FREE(m_pcC);
+	m_pcC = 0;
+	if(m_qcC) 
+		CATNET_FREE(m_qcC);
+	m_qcC = 0;
+
+	// update the betas and sigma 
+	res = findNodeMarginalProb(nnode, psamples, nsamples);
+	if(res != ERR_CATNET_OK)
+		return res;
+	estimateParameters(nnode, psamples, nsamples, m_lambdas[nnode], m_loglambdas[nnode]);
+
+	if(m_pc) 
+		CATNET_FREE(m_pc);
+	m_pc = 0;
+	if(m_qc) 
+		CATNET_FREE(m_qc);
+	m_qc = 0;
+
+	return ERR_CATNET_OK;
+}
+
+int CExpNet::setProbability(double *psamples, int nsamples) {
+
+	int *porder;
+	int k, nnode;
+
+	if(!m_loglambdas)
+		return ERR_CATNET_INIT;
+	
+	if(!psamples || nsamples < 1)
+		return ERR_CATNET_PARAM;	
+
+	porder = getOrder();
+	if(!porder)
+		return ERR_CATNET_PROC;
+
+	for(k = 0; k < m_numNodes; k++) {
+		/* work with the sample distribution of pCurNet if necessary */
+		set_sample_cache();
+
+		nnode = porder[k];
+
+		if(estimateNodeProb(nnode, m_parents[nnode], m_numParents[nnode], psamples, nsamples) != ERR_CATNET_OK)
+			break;
+
+		if(m_pCatnetSamples)
+			CATNET_FREE(m_pCatnetSamples);
+		m_nCatnetSamples = 0;
+		m_pCatnetSamples = 0; 
+	}
+
+	CATNET_FREE(porder);
+
+	return ERR_CATNET_OK;
+}
+
+void CExpNet::set_sample_cache(int becho) {
+	if(	maxParentSet() > 3 || 
+		maxCategories()*maxParentSet() > 9 || 
+		m_numNodes*maxCategories()*maxParentSet() > 256) {
+		int nsize = (int)exp(log((double)(maxCategories()))*(1+maxParentSet()))*1000;
+		if(becho)
+			Rprintf("simulate sample cache %d\n", nsize);
+		catnetSample(nsize);
+	}
+}
+
